!/*****************************************************************************/
! *
! *  Elmer, A Finite Element Software for Multiphysical Problems
! *
! *  Copyright 1st April 1995 - , CSC - IT Center for Science Ltd., Finland
! *
! *  This program is free software; you can redistribute it and/or
! *  modify it under the terms of the GNU General Public License
! *  as published by the Free Software Foundation; either version 2
! *  of the License, or (at your option) any later version.
! *
! *  This program is distributed in the hope that it will be useful,
! *  but WITHOUT ANY WARRANTY; without even the implied warranty of
! *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! *  GNU General Public License for more details.
! *
! *  You should have received a copy of the GNU General Public License
! *  along with this program (in file fem/GPL-2); if not, write to the
! *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
! *  Boston, MA 02110-1301, USA.
! *
! *****************************************************************************/
!
!/******************************************************************************
! *
! *  Authors: Gabriel Alves Mendonçca
! *  Email:   gforti@gmail.com
! *
! *  v1 - works with steady state fields (case_coil_v3.sif)
! *  v2 - modified version to work with harmonic fields (case_coil_v3_h.sif)
! *  v3 - trying improvments in region and material definitions (case_coil_v3_h.sif)
! *
! *****************************************************************************/

!------------------------------------------------------------------------------
!> Module for reading and processing losses computed by electromagnetic solvers.
!> \ingroup Solvers
!------------------------------------------------------------------------------
SUBROUTINE OutputGam( Model,Solver,dt,Transient )
!------------------------------------------------------------------------------
  USE DefUtils
  IMPLICIT NONE
!------------------------------------------------------------------------------
  TYPE(Solver_t) :: Solver
  TYPE(Model_t) :: Model
  REAL(KIND=dp) :: dt
  LOGICAL :: Transient
!------------------------------------------------------------------------------
! Local variables
!------------------------------------------------------------------------------
  TYPE(Mesh_t), POINTER :: Mesh
  TYPE(Nodes_t), SAVE :: Nodes
  INTEGER :: Visit = 0, nF, i, j, m, n, l, dim, DGsize, tdofs, ind(3)
  LOGICAL :: Found, HarmLoss
  REAL(KIND=dp) :: activeElemArea, totLoss, tempBT, tempPs, freq

  TYPE(Variable_t),POINTER :: VarElemArea, VarBT, VarLossW, TmpVar, TargetVar, Bx
  TYPE(ValueList_t), POINTER :: Material, SolverParams
  TYPE(Element_t), POINTER :: Element
  REAL(KIND=dp), POINTER :: TmpArray(:,:), omega(:,:), a_len(:,:), TesteReal(:), &
                            Density(:), K_loss(:), alpha_loss(:)
  CHARACTER(LEN=15) :: coeffStr

  TYPE(Solver_t), POINTER :: pSolver

  SAVE Visit, Mesh

  Mesh => Solver % Mesh
  dim = CoordinateSystemDimension()
  SolverParams => GetSolverParams()

  omega => ListGetConstRealArray( Model % Simulation,'Angular Frequency',Found )
  a_len => ListGetConstRealArray( Model % Simulation,'Circuit Model Depth',Found )

  freq = omega(1,1) / (2*PI)

  VarBT => VariableGet( Mesh % Variables, 'BT', ThisOnly = .TRUE. )
  IF(.NOT. ASSOCIATED( VarBT ) ) THEN
    CALL Fatal('OutputGam','Variable BT does not exist in Elmer mesh!')
  END IF

  VarElemArea => VariableGet( Mesh % Variables, 'ElemArea', ThisOnly = .TRUE. )
  IF(.NOT. ASSOCIATED( VarElemArea ) ) THEN
    CALL Fatal('OutputGam','Variable ElemArea does not exist in Elmer mesh!')
  END IF
  
  VarLossW => VariableGet( Mesh % Variables, 'LossW', ThisOnly = .TRUE. )
  IF(.NOT. ASSOCIATED( VarLossW ) ) THEN
    CALL Fatal('OutputGam','Variable LossW does not exist in Elmer mesh!')
  END IF
  
  !  Didn't work like ProcessFields
  ! HarmLoss = ListGetLogical( SolverParams, 'Calculate Harmonic Loss', Found ) 

  totLoss = 0.0d0

  DO i = 1,GetNOFActive()

    ! Element => Mesh % Elements(i)  ! This didn't work with GetElementNodes( Nodes )
    Element => GetActiveElement(i)
    Material => GetMaterial( Element )

    DGsize = SIZE (Element % DGIndexes)

    ALLOCATE ( Density(DGsize) )
    Density = GetReal( Material, 'Density', Found, UElement = Element )
    IF (.NOT. Found) THEN
      Density(1:DGsize) = 0.0d0
    END IF

    ALLOCATE ( K_loss(DGsize) )
    K_loss = GetReal( Material, 'Steinmetz coefficient K_loss', Found, UElement = Element )
    IF (.NOT. Found) THEN
      K_loss(1:DGsize) = 0.0d0
    END IF

    ALLOCATE ( alpha_loss(DGsize) )
    alpha_loss = GetReal( Material, 'Steinmetz power alpha_loss', Found, UElement = Element )
    IF (.NOT. Found) THEN
      alpha_loss(1:DGsize) = 0.0d0
    END IF    
    
    tempBT = 0.0d0

    DO m = 1,dim
      TmpVar => VariableGet( Mesh % Variables, 'Magnetic Flux Density re E '//TRIM(i2s(m)), ThisOnly = .TRUE. )
      tempBT = tempBT + TmpVar % Values(Element % DGIndexes(1)) * TmpVar % Values(Element % DGIndexes(1))
      TmpVar => VariableGet( Mesh % Variables, 'Magnetic Flux Density im E '//TRIM(i2s(m)), ThisOnly = .TRUE. )
      tempBT = tempBT + TmpVar % Values(Element % DGIndexes(1)) * TmpVar % Values(Element % DGIndexes(1))
    END DO
    tempBT = SQRT(tempBT)

    DO l = 1, DGsize
      VarBT % Values(Element % DGIndexes(l))  = tempBT
    END DO

    !  -- Calculating element area --
    CALL GetElementNodes( Nodes )
    activeElemArea = 0.0d0

    DO n = 1, 2
      activeElemArea = activeElemArea + 0.5d0*((Nodes % x(n) * Nodes % y(n + 1)) - & 
                                              (Nodes % y(n) * Nodes % x(n + 1)))
    END DO
    activeElemArea = activeElemArea + 0.5d0*((Nodes % x(3) * Nodes % y(1)) - & 
                                            (Nodes % y(3) * Nodes % x(1)))


    ! -- Calculating the local losses (W) -- 

    tempPs = K_loss(1)*freq*tempBT**(alpha_loss(1))*activeElemArea*a_len(1,1)*Density(1)
    
    
    totLoss = totLoss + tempPs
     
    VarElemArea % Values(Element % DGIndexes(1:DGsize)) = activeElemArea
    VarBT % Values(Element % DGIndexes(1:DGsize))  = tempBT
    VarLossW % Values(Element % DGIndexes(1:DGsize)) = tempPs

    ! -- work same as above
    ! DO l = 1, DGsize
    !   VarElemArea % Values(Element % DGIndexes(l)) = activeElemArea
    !   VarBT % Values(Element % DGIndexes(l))  = tempBT
    !   VarLossW % Values(Element % DGIndexes(l)) = tempPs
    ! END DO

  END DO
  
  PRINT *, "Total loss =", totLoss

!------------------------------------------------------------------------------
END SUBROUTINE OutputGam
!------------------------------------------------------------------------------
